package org.szu.mnist;

import de.learnlib.algorithms.discriminationtree.dfa.DTLearnerDFABuilder;
import de.learnlib.algorithms.kv.dfa.KearnsVaziraniDFABuilder;
import de.learnlib.algorithms.lstar.dfa.ClassicLStarDFABuilder;
import de.learnlib.algorithms.nlstar.NLStarLearner;
import de.learnlib.algorithms.nlstar.NLStarLearnerBuilder;
import de.learnlib.algorithms.ttt.dfa.TTTLearnerDFABuilder;
import de.learnlib.api.algorithm.LearningAlgorithm;
import de.learnlib.filter.cache.dfa.DFACacheOracle;
import de.learnlib.filter.statistic.oracle.CounterOracle;
import de.learnlib.util.Experiment;
import de.learnlib.util.nfa.NFALearnerWrapper;
import de.learnlib.util.statistics.SimpleProfiler;
import net.automatalib.automata.fsa.DFA;
import net.automatalib.incremental.ConflictException;
import net.automatalib.serialization.taf.writer.TAFWriter;
import net.automatalib.words.Alphabet;
import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.szu.mnist.dfa.oracle.EvalutionOracle;
import org.szu.mnist.dfa.oracle.MyDFAOracle;
import org.szu.mnist.dfa.oracle.MySimulator;
import org.szu.mnist.dfa.oracle.MySimulatorOracle;
import org.szu.mnist.mapper.AlphaBeta;
import org.szu.mnist.sul.Model;

import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

public class DFALearner<I> {
    private  Logger log = LoggerFactory.getLogger(DFALearner.class);
    private  String MNIST_ROOT = FilenameUtils.concat(System.getProperty("user.home"),"MNIST");

    //algorithm kinds
    private  int algoKind = 0;

    //Alphabet from data
    private  boolean withDataAlphabet = true;
    //
    private  boolean withCache = true;
    private  boolean withTest = true;

    //membership query
    private  boolean withWord = false; //
    //limited number for a list of a word
    private  int listLimited = 99;
    //equivalence query
    private  boolean withPassiveE = true; //
    //equivalence query with training set
    private  int testRatio = 4; //
    //Passive DFA
    private  int pRatio = 1; // [1,10]
    private  int nRatio = 9; // [1,9]
    //equivalence query with passive DFA
    private  int eqKind = 0;
    private  int bound = 1000;
    private  int dataTime = 1;
    private  int maxRefined = 20000;

    //
    private Model sulModel;
    private AlphaBeta<I> mapper;
    private String fileName = null;

    private String DFAFile = null;
    private String ADFAFile = null;
    private boolean savingDFA = false;

    private boolean withException = false;
    private boolean withLargeSym = false;

    public DFALearner(Model sulModel, AlphaBeta<I> mapper, String fileName){
        this.sulModel = sulModel;
        this.mapper = mapper;
        this.fileName = fileName;
        if(this.fileName == null) {
            SimpleDateFormat ft = new SimpleDateFormat ("-yyMMdd-HHmmss");
            Date date = new Date();
            String sDate = ft.format(date);
            this.fileName = "temp" + sDate + ".txt";
        }
        this.DFAFile = this.fileName.substring(0,this.fileName.length()-4)+"dfa.txt";
        this.ADFAFile = this.fileName.substring(0,this.fileName.length()-4)+"adfa.txt";
    }


    public void setAlgoKind(int algoKind) {
        this.algoKind = algoKind;
    }

    public void setWithDataAlphabet(boolean withDataAlphabet) {
        this.withDataAlphabet = withDataAlphabet;
    }

    public void setWithCache(boolean withCache) {
        this.withCache = withCache;
    }

    public void setWithTest(boolean withTest) {
        this.withTest = withTest;
    }

    public void setWithWord(boolean withWord) {
        this.withWord = withWord;
    }

    public void setListLimited(int listLimited) {
        this.listLimited = listLimited;
    }

    public void setWithPassiveE(boolean withPassiveE) {
        this.withPassiveE = withPassiveE;
    }

    public void setTestRatio(int testRatio) {
        this.testRatio = testRatio;
    }

    public void setpRatio(int pRatio) {
        this.pRatio = pRatio;
    }

    public void setnRatio(int nRatio) {
        this.nRatio = nRatio;
    }

    public void setEqKind(int eqKind) {
        this.eqKind = eqKind;
    }

    public void setBound(int bound) {
        this.bound = bound;
    }

    public void setDataTime(int dataTime) {
        this.dataTime = dataTime;
    }

    public void setMaxRefined(int maxRefined) {
        this.maxRefined = maxRefined;
    }

    public void setSavingDFA(boolean savingDFA) {
        this.savingDFA = savingDFA;
    }

    public void learning() {
        log.info("Get the input alphabet of system under learning ...");
        Alphabet<I> inputs = mapper.getAlphabetFromData();
        System.out.println("Sigma: " + inputs.size());
        if(!withDataAlphabet) inputs = mapper.getAlphabet();
        if(withDataAlphabet) {
            System.out.println("Sigma indeed: " + mapper.getAlphabetSize().toString());
        }

        log.info("Building simulator ...");
        MySimulator<I> simulator = new MySimulator<>(mapper);
        simulator.setpRadio(pRatio);
        simulator.setnRadio(nRatio);
        DFA<?, I> pdfa = simulator.getModel();

        // oracle for counting queries wraps SUL
        log.info("Building oracle for membership and equivalence query ...");
        MySimulatorOracle<I> myOracle = new MySimulatorOracle<>(mapper,simulator);
        //for membership
        myOracle.setWithWord(withWord);
        //for equivalence with passive
        myOracle.setWithPassiveE(withPassiveE);
        myOracle.setEqKind(eqKind);
        myOracle.setBound(bound);
        myOracle.setDataTime(dataTime);
        myOracle.setMaxRefind(maxRefined);
        //for equivalence with training dataset
        myOracle.setTestingRatio(testRatio);

        this.withLargeSym = this.mapper.isLargeSym();


        CounterOracle.DFACounterOracle<I> cmqOracle = new CounterOracle.DFACounterOracle<>(myOracle,"membership");
        DFACacheOracle<I> mqCacheOracle = DFACacheOracle.createDAGCacheOracle(inputs,cmqOracle);

        // construct L* instance
        log.info("Setting the learning algorithm  ...");
        //
        LearningAlgorithm<? extends DFA<?, I>, I, Boolean> learningAlgorithm;

        switch(algoKind) {
            case 0 :
                if(withCache) learningAlgorithm = new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(mqCacheOracle) // membership oracle
                    .create();
                else learningAlgorithm =  new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(cmqOracle) // membership oracle
                    .create();
                break;
            case 1:
                if(withCache) learningAlgorithm = new DTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(mqCacheOracle) // membership oracle
                    .create();
                else learningAlgorithm =  new DTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(cmqOracle) // membership oracle
                    .create();
                break;
            case 2:
                if(withCache) learningAlgorithm = new KearnsVaziraniDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(mqCacheOracle) // membership oracle
                    .create();
                else learningAlgorithm =  new KearnsVaziraniDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(cmqOracle) // membership oracle
                    .create();
                break;
            case 3:
                if(withCache) learningAlgorithm = new TTTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(mqCacheOracle) // membership oracle
                    .create();
                else learningAlgorithm =  new TTTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(cmqOracle) // membership oracle
                    .create();
                break;
            case 4:
                // construct NL* instance
                NLStarLearner<I> nlstar;
                if(withCache) nlstar = new NLStarLearnerBuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(mqCacheOracle) // membership oracle
                    .create();
                else nlstar =  new NLStarLearnerBuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(cmqOracle) // membership oracle
                    .create();
                // NL* wrapper
                learningAlgorithm = new NFALearnerWrapper(inputs, nlstar);
                break;
            default:
                if(withCache) learningAlgorithm = new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(mqCacheOracle) // membership oracle
                    .create();
                else learningAlgorithm =  new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                    .withOracle(cmqOracle) // membership oracle
                    .create();
                break;
        }

        /*DefaultQuery<Integer, Boolean> counterexample = null;
        int round = 0;
        do {
            round++;
            if (counterexample == null) {
                learningAlgorithm.startLearning();
            } else {
                boolean refined = learningAlgorithm.refineHypothesis(counterexample);
                if (!refined) {
                    System.err.println("No refinement effected by counterexample!");
                    break;
                }
            }
            log.info("Round " + round);
            DFA<?, Integer> result = learningAlgorithm.getHypothesisModel();
            DFAWriter.printDFA(result,inputs,"learnDFA"+round+".txt");
            counterexample = eqOracle.findCounterExample(learningAlgorithm.getHypothesisModel(), inputs);
            if(counterexample != null) log.info(counterexample.toString());
        } while (counterexample != null); */

        //Experiment
        log.info("Setting the experiment ...");
        Experiment.DFAExperiment<I> experiment = new Experiment.DFAExperiment<>(learningAlgorithm, myOracle, inputs);
        // turn on time profiling
        experiment.setProfile(true);
        // enable logging of models
        experiment.setLogModels(true);
        // run experiment
        log.info("Running the experiment ...");
        long startTime1 = 0;
        long startTime = System.currentTimeMillis();
        try {
            experiment.run();

        } catch (Exception e){ //ConflictException ec
            System.out.println("===========Exception: " + e.toString() + "==========");
            this.withException = true;
            //System.out.println("======================================BEGINBEGINBEGINBEGINBEGINBEGINBEGINBEGIN=======================================");
            MyDFAOracle<I> myDFAOracle = new MyDFAOracle<>(mapper, myOracle.getAbstractDFA());
            myDFAOracle.setEqKind(eqKind);
            myDFAOracle.setBound(bound);
            cmqOracle = new CounterOracle.DFACounterOracle<>(myDFAOracle,"membership");

            //mqCacheOracle = DFACacheOracle.createDAGCacheOracle(inputs,cmqOracle);
            switch(algoKind) {
                case 0 :
                    /*if(withCache) learningAlgorithm = new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    learningAlgorithm =  new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
                case 1:
                    /*if(withCache) learningAlgorithm = new DTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    learningAlgorithm =  new DTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
                case 2:
                    /*if(withCache) learningAlgorithm = new KearnsVaziraniDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    learningAlgorithm =  new KearnsVaziraniDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
                case 3:
                    /*if(withCache) learningAlgorithm = new TTTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                     learningAlgorithm =  new TTTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
                case 4:
                    // construct NL* instance
                    NLStarLearner<I> nlstar;
                    /*if(withCache) nlstar = new NLStarLearnerBuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    nlstar =  new NLStarLearnerBuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    // NL* wrapper
                    learningAlgorithm = new NFALearnerWrapper(inputs, nlstar);
                    break;
                default:
                    /*if(withCache) learningAlgorithm = new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    learningAlgorithm =  new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
            }
            experiment = new Experiment.DFAExperiment<>(learningAlgorithm, myDFAOracle, inputs);
            // turn on time profiling
            experiment.setProfile(true);
            // enable logging of models
            experiment.setLogModels(true);
            startTime1 = System.currentTimeMillis();
            experiment.run();
            //System.out.println("======================================ENDENDENDENDENDENDENDENDENDENDEND=======================================");
        }

        //The rounds for learning is too fewer to get a fine dfa. Then to learn the dfa from the abstracted DFA.
        if(!this.withLargeSym && !this.withException && experiment.getRounds().getCount() <= 50) {
            System.out.println("===========Exception: the rounds for learning is too fewer and then learn from the abstracted DFA directly==========");
            this.withException = true;
            //System.out.println("======================================BEGINBEGINBEGINBEGINBEGINBEGINBEGINBEGIN=======================================");
            MyDFAOracle<I> myDFAOracle = new MyDFAOracle<>(mapper, myOracle.getAbstractDFA());
            myDFAOracle.setEqKind(eqKind);
            myDFAOracle.setBound(bound);
            cmqOracle = new CounterOracle.DFACounterOracle<>(myDFAOracle,"membership");

            //mqCacheOracle = DFACacheOracle.createDAGCacheOracle(inputs,cmqOracle);
            switch(algoKind) {
                case 0 :
                    /*if(withCache) learningAlgorithm = new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    learningAlgorithm =  new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
                case 1:
                    /*if(withCache) learningAlgorithm = new DTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    learningAlgorithm =  new DTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
                case 2:
                    /*if(withCache) learningAlgorithm = new KearnsVaziraniDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    learningAlgorithm =  new KearnsVaziraniDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
                case 3:
                    /*if(withCache) learningAlgorithm = new TTTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    learningAlgorithm =  new TTTLearnerDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
                case 4:
                    // construct NL* instance
                    NLStarLearner<I> nlstar;
                    /*if(withCache) nlstar = new NLStarLearnerBuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    nlstar =  new NLStarLearnerBuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    // NL* wrapper
                    learningAlgorithm = new NFALearnerWrapper(inputs, nlstar);
                    break;
                default:
                    /*if(withCache) learningAlgorithm = new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(mqCacheOracle) // membership oracle
                        .create();
                    else*/
                    learningAlgorithm =  new ClassicLStarDFABuilder<I>().withAlphabet(inputs) // input alphabet
                        .withOracle(cmqOracle) // membership oracle
                        .create();
                    break;
            }
            experiment = new Experiment.DFAExperiment<>(learningAlgorithm, myDFAOracle, inputs);
            // turn on time profiling
            experiment.setProfile(true);
            // enable logging of models
            experiment.setLogModels(true);
            startTime1 = System.currentTimeMillis();
            experiment.run();
            //System.out.println("======================================ENDENDENDENDENDENDENDENDENDENDEND=======================================");
        }


        // get learned model
        log.info("Load the learned results ...");
        DFA<?, I> result = experiment.getFinalHypothesis();
        long endTime = System.currentTimeMillis();

        try {
            FileOutputStream fos = new FileOutputStream(fileName);
            PrintWriter out = new PrintWriter(fos);
            out.println("algoKind: " + algoKind);
          //  out.println("withWord: " + withWord);
         //   out.println("withPassiveE: " + withPassiveE);
          //  out.println("listLimited: " + listLimited);
           // out.println("pRatio: " + pRatio);
           // out.println("nRatio: " + nRatio);
           // out.println("withException: " + withException);
            if (withPassiveE) {
                out.println("eqKind: " + eqKind);
              //  out.println("bound: " + bound);
               // out.println("dataTime: " + dataTime);
                out.println("maxRefined: " + maxRefined);
            } else {
                out.println("testRatio: " + testRatio);
            }

            // report results
            System.out.println("-------------------------------------------------------");
            System.out.println("Learning Time: " + (endTime - startTime) + "ms");
            out.println("Learning Time: " + (endTime - startTime) + "ms");
            /*
            if(withException) {
                System.out.println("ELearning Time with Exception: " + (endTime - startTime1) + "ms");
                out.println("ELearning Time with Exception: " + (endTime - startTime1) + "ms");
            }*/
            // profiling
            System.out.println(SimpleProfiler.getResults());
            out.println(SimpleProfiler.getResults());
            // learning statistics
            System.out.println(experiment.getRounds().getSummary());
            out.println(experiment.getRounds().getSummary());
            System.out.println(cmqOracle.getStatisticalData().getSummary());
            out.println(cmqOracle.getStatisticalData().getSummary());
            if (withPassiveE) {
                System.out.println("Refine: " + myOracle.getRefimeTime());
                out.println("Refine: " + myOracle.getRefimeTime());
            }

            // model statistics
            System.out.println("States: " + result.size());
            out.println("States: " + result.size());
            System.out.println("Sigma: " + inputs.size());
            out.println("Sigma: " + inputs.size());
            /*
            if (withPassiveE) {
                //System.out.println("States0: " + pdfa.size());
                //out.println("States0: " + pdfa.size());
                System.out.println("States1: " + simulator.getModel().size());
                out.println("States1: " + simulator.getModel().size());
            }*/


            if (withTest) {
                EvalutionOracle<I> tester = new EvalutionOracle<>(out);
                log.info("Testing the learned dfa with training data ...");
                out.println("Testing the learned dfa with training data ...");
                tester.eval(result, mapper, true);
                tester.evalByData(result, mapper, true);
                log.info("Testing the learned dfa with testing data ...");
                out.println("Testing the learned dfa with testing data ...");
                tester.eval(result, mapper, false);
                tester.evalByData(result, mapper, false);

                if (withPassiveE) {
                    /*out.println("Testing the passive dfa with training data ...");
                    log.info("Testing the passive dfa with training data ...");
                    tester.eval(pdfa, mapper, true);
                    tester.evalByData(pdfa, mapper, true);
                    log.info("Testing the passive dfa with testing data ...");
                    out.println("Testing the passive dfa with testing data ...");
                    tester.eval(pdfa, mapper, false);
                    tester.evalByData(pdfa, mapper, false);*/


                    log.info("Testing the abstracted dfa with training data ...");
                    out.println("Testing the abstracted dfa with training data ...");
                    tester.eval(simulator.getModel(), mapper, true);
                    tester.evalByData(simulator.getModel(), mapper, true);
                    log.info("Testing the abstracted dfa with testing data ...");
                    out.println("Testing the abstracted dfa with testing data ...");
                    tester.eval(simulator.getModel(), mapper, false);
                    tester.evalByData(simulator.getModel(), mapper, false);


                    /*List<DFA<?, I>> models = new ArrayList<>();
                    models.add(result);
                    models.add(pdfa);
                    models.add(simulator.getModel());
                    log.info("Testing the result dfa with training data ...");
                    out.println("Testing the result dfa with training data ...");
                    tester.evals(models, mapper, true);
                    tester.evalsByData(models, mapper, true);
                    log.info("Testing the result dfa with testing data ...");
                    out.println("Testing the result dfa with testing data ...");
                    tester.evals(models, mapper, false);
                    tester.evalsByData(models, mapper, false);*/
                }
            }

            out.close();
            fos.close();

            if(this.savingDFA) {
                log.info("Saving the learned dfa ...");
                FileOutputStream fos1 = new FileOutputStream(this.DFAFile);
                PrintWriter out1 = new PrintWriter(fos1);
                TAFWriter.writeDFA(result, inputs, out1);
                out1.close();
                fos1.close();
                if (withPassiveE) {
                    log.info("Saving the abstracted dfa ...");
                    fos1 = new FileOutputStream(this.ADFAFile);
                    out1 = new PrintWriter(fos1);
                    TAFWriter.writeDFA(simulator.getModel(), inputs, out1);
                    out1.close();
                    fos1.close();
                }
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
