package org.szu.mnist.sul;

import org.apache.commons.io.FilenameUtils;
import org.deeplearning4j.eval.Evaluation;
import org.deeplearning4j.nn.conf.MultiLayerConfiguration;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.ConvolutionLayer;
import org.deeplearning4j.nn.conf.layers.DenseLayer;
import org.deeplearning4j.nn.conf.layers.OutputLayer;
import org.deeplearning4j.nn.conf.layers.SubsamplingLayer;
import org.deeplearning4j.nn.multilayer.MultiLayerNetwork;
import org.deeplearning4j.nn.weights.WeightInit;
import org.deeplearning4j.optimize.listeners.ScoreIterationListener;
import org.deeplearning4j.util.ModelSerializer;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.DataSet;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.factory.Nd4j;
import org.nd4j.linalg.learning.config.Nesterovs;
import org.nd4j.linalg.lossfunctions.LossFunctions;
import org.nd4j.linalg.schedule.MapSchedule;
import org.nd4j.linalg.schedule.ScheduleType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;

public class MnistClassifierModel extends Model {
    private static Logger log = LoggerFactory.getLogger(MnistClassifierModel.class);

    private final int numRows = DataProvider.numRows;
    private final int numColumns = DataProvider.numColumns;
    private final int outputNum = DataProvider.outputNum; // number of output classes, here we use binary classifier for simplcity
    //private int batchSize = 128; // batch size for each epoch
    private final int channels = 1; // single channel for grayscale images
    private final int rngSeed = 1234; // random number seed for reproducibility
    private final int numEpochs = 15; // number of epochs to perform


    private DataSetIterator mnistTrain = null;
    // private  DataSetIterator mnistTest = null;
    private String MNIST_ROOT = FilenameUtils.concat(System.getProperty("user.home"),"MNIST");
    private String modefileName = "MnistClassifierBModel"+ DataProvider.getKind() + ".zip";
    private String modelfile = FilenameUtils.concat(MNIST_ROOT, modefileName);
    private MultiLayerNetwork model = null;

    public MnistClassifierModel() {
        if(!modelExist() && this.model == null) buildModel();
    }

    public MnistClassifierModel(DataSetIterator mnistTrain) {
        this.mnistTrain = mnistTrain;
        buildModel();
    }

    public void setTrainData(DataSetIterator mnistTrain) {
        this.mnistTrain = mnistTrain;
    }


    private void buildModel()  {
        try {
            if(this.mnistTrain == null) {
                this.mnistTrain = DataProvider.getMnistTrainByKind();
            }

            log.info("Network configuration and training...");
            Map<Integer, Double> lrSchedule = new HashMap<>();
            lrSchedule.put(0, 0.06); // iteration #, learning rate
            lrSchedule.put(200, 0.05);
            lrSchedule.put(600, 0.028);
            lrSchedule.put(800, 0.0060);
            lrSchedule.put(1000, 0.001);

            log.info("Build model....");
            MultiLayerConfiguration conf = new NeuralNetConfiguration.Builder()
                .seed(rngSeed)
                .l2(0.0005)
                .updater(new Nesterovs(new MapSchedule(ScheduleType.ITERATION, lrSchedule)))
                .weightInit(WeightInit.XAVIER)
                .list()
                .layer(0, new ConvolutionLayer.Builder(5, 5)
                    .nIn(channels)
                    .stride(1, 1)
                    .nOut(20)
                    .activation(Activation.IDENTITY)
                    .build())
                .layer(1, new SubsamplingLayer.Builder(SubsamplingLayer.PoolingType.MAX)
                    .kernelSize(2, 2)
                    .stride(2, 2)
                    .build())
                .layer(2, new ConvolutionLayer.Builder(5, 5)
                    .stride(1, 1) // nIn need not specified in later layers
                    .nOut(50)
                    .activation(Activation.IDENTITY)
                    .build())
                .layer(3, new SubsamplingLayer.Builder(SubsamplingLayer.PoolingType.MAX)
                    .kernelSize(2, 2)
                    .stride(2, 2)
                    .build())
                .layer(4, new DenseLayer.Builder().activation(Activation.RELU)
                    .nOut(500).build())
                .layer(5, new OutputLayer.Builder(LossFunctions.LossFunction.NEGATIVELOGLIKELIHOOD)
                    .nOut(outputNum)
                    .activation(Activation.SOFTMAX)
                    .build())
                .setInputType(InputType.convolutionalFlat(28, 28, 1)) // InputType.convolutional for normal image
                .backprop(true).pretrain(false).build();

            model = new MultiLayerNetwork(conf);
            model.init();
            model.setListeners(new ScoreIterationListener(10));
            log.debug("Total num of params: {}", model.numParams());


            log.info("Train model ...");
            for( int i=0; i<numEpochs; i++ ){
                model.fit(mnistTrain);
            }

            //log.info("Evaluate model ...");
            eval(DataProvider.getMnistTestByKind());

            log.info("****SAVE THE MODEL***");
            //where to save model
            File locationToSave = new File(modelfile);
            //boolean save updater
            boolean saveUpdater = false;
            ModelSerializer.writeModel(model, locationToSave, saveUpdater);
        }catch (java.lang.Exception e) {
            log.info("[Error] Fail to save model.");
            e.printStackTrace();
        }
    }

    private boolean modelExist() {
        File saveModel = new File(modelfile);
        if (!saveModel.exists()) return false;
        if(this.model == null) {
            log.info("****LOAD THE MODEL***");
            try {
                this.model = ModelSerializer.restoreMultiLayerNetwork(saveModel);
            } catch(java.io.IOException e) {
                log.info("[Error] Fail to load model.");
                return false;
            }
        }
        return true;
    }

    public MultiLayerNetwork getModel() {
        if(!modelExist() && this.model == null) buildModel();
        return this.model;
    }

    @Override
    public Boolean query(INDArray input, INDArray label) {
        //log.info("Evaluate model....");
        INDArray guess = model.output(input); //get the networks prediction test.getFeatureMatrix()
        INDArray guessIndex = Nd4j.argMax(guess, 1);
        INDArray labelIndex = Nd4j.argMax(label, 1);
        if (guessIndex.getInt(0) == labelIndex.getInt(0)) return true;
        return false;
    }

    @Override
    public Boolean query(INDArray input) {
        //log.info("Evaluate model....");
        INDArray guess = model.output(input); //get the networks prediction test.getFeatureMatrix()
        INDArray guessIndex = Nd4j.argMax(guess, 1);
        if (guessIndex.getInt(0) > 0) return true;
        return false;
    }

    public void eval(DataSetIterator mnistTest) {
        log.info("Evaluate model....");
        Evaluation eval = new Evaluation(outputNum); //create an evaluation object with 10 possible classes
        while(mnistTest.hasNext()){
            DataSet next = mnistTest.next();
            INDArray output = model.output(next.getFeatureMatrix()); //get the networks prediction
            eval.eval(next.getLabels(), output); //check the prediction against the true class
        }
        log.info(eval.stats());
        log.info("****************Example finished********************");
        try {
            String fileName = modefileName.substring(0, modefileName.length() - 4) + ".txt";
            FileOutputStream fos = new FileOutputStream(fileName);
            PrintWriter out = new PrintWriter(fos);
            out.print(eval.stats());
            out.close();
            fos.close();
        }catch (Exception e) {
            e.printStackTrace();
        }


    }

}
